<?php
/**
 * Auth actions.
 *
 * @package Serviceez\Controllers\Auth
 */

namespace Serviceez\Controllers\Auth;

use WP_Error;
use WP_REST_Request;
use WP_REST_Response;
use WP_User;

/**
 * Auth actions.
 *
 * @package Serviceez\Controllers\Auth
 */
class Actions {
	/**
	 * Login a user
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response|WP_Error
	 */
	public function login( WP_REST_Request $request ) {
		$creds = array(
			'user_login'    => sanitize_email( $request->get_param( 'email' ) ),
			'user_password' => $request->get_param( 'password' ),
			'remember'      => (bool) $request->get_param( 'remember' ),
		);

		$user = wp_signon( $creds, false );

		if ( is_wp_error( $user ) ) {
			return new WP_Error( 'login_failed', 'Invalid credentials', array( 'status' => 401 ) );
		}

		return new WP_REST_Response(
			array(
				'success' => true,
				'message' => 'Login successful',
				'user'    => $this->get_user_data( $user ),
			),
			200
		);
	}

	/**
	 * Register a new client user
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response|WP_Error
	 */
	public function register( WP_REST_Request $request ) {
		$email    = sanitize_email( $request->get_param( 'email' ) );
		$password = $request->get_param( 'password' );
		$name     = sanitize_text_field( $request->get_param( 'name' ) );

		if ( \email_exists( $email ) ) {
			return new WP_Error( 'registration_failed', 'Email already exists', array( 'status' => 400 ) );
		}
		$user_id = \wp_create_user( $email, $password, $email );

		if ( \is_wp_error( $user_id ) ) {
			return new WP_Error( 'registration_failed', $user_id->get_error_message(), array( 'status' => 400 ) );
		}

		// Set the user role
		$user = new \WP_User( $user_id );
		$user->set_role( 'serviceez_client' );

		\wp_update_user(
			array(
				'ID'         => $user_id,
				'first_name' => $name,
			)
		);

		// Fetch the updated user data.
		$updated_user = get_user_by( 'id', $user_id );

		// Ensure the role is set
		if ( empty( $updated_user->roles ) ) {
			$updated_user->add_role( 'serviceez_client' );
		}

		return new WP_REST_Response(
			array(
				'success' => true,
				'message' => 'Registration successful',
				'user'    => $this->get_user_data( $updated_user ),
			),
			201
		);
	}

	/**
	 * Get user data
	 *
	 * @param WP_User $user The user object.
	 * @return array
	 */
	private function get_user_data( $user ) {
		$roles = $user->roles;
		$role  = ! empty( $roles ) ? $roles[0] : 'None';

		return array(
			'id'    => $user->ID,
			'email' => $user->user_email,
			'name'  => $user->first_name,
			'role'  => $role,
		);
	}

	/**
	 * Check nonce.
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_Error|true
	 */
	public function check( WP_REST_Request $request ) {
		// check nonce.
		$nonce = $request->get_header( 'X-WP-Nonce' );

		if ( ! $nonce ) {
			return new WP_Error( 'invalid_nonce', 'Invalid nonce', array( 'status' => 401 ) );
		}

		if ( ! wp_verify_nonce( $nonce, 'wp_rest' ) ) {
			return new WP_Error( 'invalid_nonce', 'Invalid nonce', array( 'status' => 401 ) );
		}

		return true;
	}
}
