<?php
/**
 * Coupon actions.
 *
 * @package Serviceez\Controllers\Coupon
 */

namespace Serviceez\Controllers\Coupon;

use ServiceEZ\Models\Coupon;

/**
 * Coupon actions.
 *
 * @package Serviceez\Controllers\Coupon
 */
class Actions {
	/**
	 * Get all coupons.
	 *
	 * @return array The coupons.
	 */
	public function index() {
		return Coupon::all();
	}

	/**
	 * Store a coupon.
	 *
	 * @param \WP_REST_Request $request The request object.
	 * @return \WP_REST_Response The response object.
	 */
	public function store( \WP_REST_Request $request ) {
		$valid_until = null;
		if ( 'custom' === $request['validity_type'] && ! empty( $request['valid_until'] ) ) {
			$valid_until = gmdate( 'Y-m-d H:i:s', strtotime( $request['valid_until'] ) );
		}

		$coupon = Coupon::create(
			array(
				'code'          => $request['code'],
				'discount'      => $request['discount'],
				'type'          => $request['type'],
				'validity_type' => $request['validity_type'],
				'valid_until'   => $valid_until,
			)
		);

		return $coupon;
	}

	/**
	 * Update a coupon.
	 *
	 * @param \WP_REST_Request $request The request object.
	 * @return \WP_REST_Response|\WP_Error The response object or error.
	 */
	public function update( \WP_REST_Request $request ) {
		$id     = $request['id'];
		$coupon = Coupon::where( 'id', $id )->first();

		if ( ! $coupon ) {
			return new \WP_Error( 'not_found', 'Coupon not found', array( 'status' => 404 ) );
		}

		$valid_until = $coupon->valid_until;
		if ( 'custom' === $request['validity_type'] && ! empty( $request['valid_until'] ) ) {
			$valid_until = gmdate( 'Y-m-d H:i:s', strtotime( $request['valid_until'] ) );
		} elseif ( 'forever' === $request['validity_type'] ) {
			$valid_until = null;
		}

		$coupon->update(
			array(
				'code'          => $request['code'] ?? $coupon->code,
				'discount'      => $request['discount'] ?? $coupon->discount,
				'type'          => $request['type'] ?? $coupon->type,
				'validity_type' => $request['validity_type'] ?? $coupon->validity_type,
				'valid_until'   => $valid_until,
			)
		);

		return $coupon;
	}

	/**
	 * Delete a coupon.
	 *
	 * @param \WP_REST_Request $request The request object.
	 * @return \WP_REST_Response|\WP_Error The response object or error.
	 */
	public function delete( \WP_REST_Request $request ) {
		$id     = $request['id'];
		$coupon = Coupon::find( $id );

		if ( ! $coupon ) {
			return new \WP_Error( 'not_found', 'Coupon not found', array( 'status' => 404 ) );
		}

		$coupon->delete();

		return array( 'message' => 'Coupon deleted successfully' );
	}

	/**
	 * Verify a coupon.
	 *
	 * @param \WP_REST_Request $request The request object.
	 * @return \WP_REST_Response|\WP_Error The response object or error.
	 */
	public function verify( \WP_REST_Request $request ) {
		$code   = $request['code'];
		$coupon = Coupon::where( 'code', $code )->first();

		if ( ! $coupon ) {
			return new \WP_Error( 'not_found', 'Coupon not found', array( 'status' => 404 ) );
		}

		// Check if the coupon is valid (not expired).
		if ( 'custom' === $coupon->validity_type && $coupon->valid_until < current_time( 'mysql' ) ) {
			return new \WP_Error( 'expired', 'Coupon has expired', array( 'status' => 400 ) );
		}

		return array(
			'id'            => $coupon->id,
			'code'          => $coupon->code,
			'discount'      => $coupon->discount,
			'type'          => $coupon->type,
			'validity_type' => $coupon->validity_type,
			'valid_until'   => $coupon->valid_until,
		);
	}
}
