<?php

/**
 * Order Actions Controller
 *
 * This file contains the Actions class for handling Order-related operations.
 *
 * @package ServiceEZ
 * @subpackage Controllers\Orders
 */

namespace ServiceEZ\Controllers\Orders;

use ServiceEZ\Core\Template;
use ServiceEZ\Models\Order;
use ServiceEZ\Models\Client;
use WP_REST_Request;
use WP_REST_Response;
use ServiceEZ\Models\Service;
use ServiceEZ\Helpers\EmailHelper;
use Stripe\Stripe;
use Stripe\PaymentIntent;
use ServiceEZ\Models\Notification;
use ServiceEZ\Models\Users;

/**
 * Class Actions
 *
 * Handles CRUD operations for Orders.
 */
class Actions {

	/**
	 * Create a new order.
	 *
	 * @param WP_REST_Request $request The incoming request object.
	 * @return WP_REST_Response The response containing the created order.
	 */
	public function create( $request ) {
		$params = $request->get_params();

		// Check if the service exists.
		$service = Service::where( 'id', $params['service_id'] )->first();

		if ( ! $service ) {
			return new WP_REST_Response( array( 'message' => 'Invalid service ID' ), 400 );
		}

		// Check if a client with the given email already exists.
		$client = Client::where( 'email', $params['email'] ?? '' )->first();

		if ( ! $client ) {
			$user_id = $this->create_wordpress_user( $params );

			if ( $user_id ) {
				// If client doesn't exist, create a new one.
				$client = Client::create(
					array(
						'user_id'       => $user_id,
						'firstName'     => $params['firstName'] ?? '',
						'lastName'      => $params['lastName'] ?? '',
						'email'         => $params['email'] ?? '',
						'address_line1' => $params['address']['line1'] ?? '',
						'address_line2' => $params['address']['line2'] ?? '',
						'city'          => $params['address']['city'] ?? '',
						'state'         => $params['address']['state'] ?? '',
						'postal_code'   => $params['address']['postalCode'] ?? '',
						'country'       => $params['address']['country'] ?? '',
					// Add other client fields as necessary.
					)
				);

				// Create notification.
				$this->create_notification(
					// translators: %s: Client name.
					sprintf( __( 'New client %s created', 'serviceez' ), $client->name ),
					'client_created',
					"/clients/{$client->id}"
				);
			}
		} else {
			$user_id = $this->create_wordpress_user( $params );

			// Update the client.
			Client::where( 'id', $client->id )->update( array( 'user_id' => $user_id ) );
		}

		// Prepare the order data.
		$order_data = isset( $params['order_data'] ) ? json_decode( $params['order_data'], true ) : array();

		// Ensure $order_data is an array.
		$order_data = is_array( $order_data ) ? $order_data : array();

		// Add tax information to order_summary.
		if ( ! isset( $order_data['order_summary'] ) ) {
			$order_data['order_summary'] = array();
		}

		// Use the tax and taxRate from the params, or fallback to the values in order_data if they exist.
		$order_data['order_summary']['tax']     = isset( $params['tax'] ) ? $params['tax'] : ( $order_data['order_summary']['tax'] ?? 0 );
		$order_data['order_summary']['taxRate'] = isset( $params['taxRate'] ) ? $params['taxRate'] : ( $order_data['order_summary']['taxRate'] ?? 0 );

		// Create the order associated with the client.
		try {
			$order = Order::create(
				array(
					'client_id'      => $client->id,
					'service_id'     => $service->id,
					'order_status'   => $params['order_status'] ?? 'pending',
					'payment_status' => $params['payment_status'] ?? 'pending',
					'payment_method' => $params['payment_method'] ?? 'none',
					'total_amount'   => $params['total_amount'] ?? 0,
					'data'           => wp_json_encode( $order_data ),
				)
			);

			// Create notification.
			$this->create_notification(
				// translators: %1$d: Order ID, %2$s: Client name.
				sprintf( __( 'New order #%1$d created for %2$s', 'serviceez' ), $order->id, $client->name ),
				'order_created',
				"/orders/{$order->id}"
			);

			// Send invoice email.
			$this->send_invoice_email( $order );

			return new WP_REST_Response( $order, 201 );
		} catch ( \Exception $e ) {

			return new WP_REST_Response( array( 'message' => 'Error creating order' ), 500 );
		}
	}

	/**
	 * Send the invoice email.
	 *
	 * @param object $order The order object.
	 */
	private function send_invoice_email( $order ) {
		$client       = $order->client;
		$order_data   = json_decode( $order->data, true );
		$invoice_link = Template::get_template_slug_used_on_page()[0]->guid . '#/orders/invoice/' . $order->id;

		$email_content = $this->generate_invoice_email_content( $order, $invoice_link );

		EmailHelper::send_email(
			$client->email,
			'Your Invoice for Order #' . $order->id,
			$email_content
		);
	}

	/**
	 * Create a WordPress user for the client.
	 *
	 * @param array $params The parameters array.
	 */
	private function create_wordpress_user( $params ) {
		// Check if the user already exists.
		$user = get_user_by( 'email', $params['email'] );
		if ( $user ) {
			return $user->ID;
		}

		// Create the user.
		$user_id = wp_create_user( $params['email'], wp_generate_password(), $params['email'] );

		// Check for errors.
		if ( is_wp_error( $user_id ) ) {
			return null;
		}

		return $user_id;
	}

	/**
	 * Generate the invoice email content.
	 *
	 * @param object $order The order object.
	 * @param string $invoice_link The invoice link.
	 * @return string The invoice email content.
	 */
	private function generate_invoice_email_content( $order, $invoice_link ) {
		$order_data    = json_decode( $order->data, true );
		$order_summary = $order_data['order_summary'];

		ob_start();
		?>
		<!DOCTYPE html>
		<html lang="en">

		<head>
			<meta charset="UTF-8">
			<meta name="viewport" content="width=device-width, initial-scale=1.0">
			<title>
		<?php
			// translators: %d: Order ID.
			echo esc_html( sprintf( __( 'Invoice for Order #%d', 'serviceez' ), $order->id ) );
		?>
			</title>
			<style>
				body {
					font-family: Arial, sans-serif;
					line-height: 1.6;
					color: #333;
				}

				.container {
					max-width: 600px;
					margin: 0 auto;
					padding: 20px;
				}

				.header {
					text-align: center;
					margin-bottom: 20px;
				}

				.invoice-details {
					margin-bottom: 20px;
				}

				.invoice-items {
					width: 100%;
					border-collapse: collapse;
					margin-bottom: 20px;
				}

				.invoice-items th,
				.invoice-items td {
					border: 1px solid #ddd;
					padding: 10px;
					text-align: left;
				}

				.total {
					font-weight: bold;
				}

				.button {
					display: inline-block;
					padding: 10px 20px;
					background-color: #007bff;
					color: #ffffff;
					text-decoration: none;
					border-radius: 5px;
				}
			</style>
		</head>

		<body>
			<div class="container">
				<div class="header">
					
					<h1>
				<?php
					// translators: %d: Order ID.
					echo esc_html( sprintf( __( 'Invoice for Order #%d', 'serviceez' ), $order->id ) );
				?>
					</h1>
				</div>
				<div class="invoice-details">
					<p><strong><?php esc_html_e( 'Date:', 'serviceez' ); ?></strong> <?php echo esc_html( gmdate( 'F j, Y', strtotime( $order->created_at ) ) ); ?></p>
					<p><strong><?php esc_html_e( 'Client:', 'serviceez' ); ?></strong> <?php echo esc_html( $order->client->name ); ?></p>
					<p><strong><?php esc_html_e( 'Amount Due:', 'serviceez' ); ?></strong> $<?php echo esc_html( number_format( $order_summary['total'], 2 ) ); ?></p>
				</div>
				<table class="invoice-items">
					<thead>
						<tr>
							<th><?php esc_html_e( 'Item', 'serviceez' ); ?></th>
							<th><?php esc_html_e( 'Quantity', 'serviceez' ); ?></th>
							<th><?php esc_html_e( 'Price', 'serviceez' ); ?></th>
							<th><?php esc_html_e( 'Total', 'serviceez' ); ?></th>
						</tr>
					</thead>
					<tbody>
						<tr>
							<td><?php echo esc_html( $order_summary['package']['name'] ); ?></td>
							<td><?php echo esc_html( $order_summary['package']['quantity'] ); ?></td>
							<td>$<?php echo esc_html( number_format( $order_summary['package']['total'], 2 ) ); ?></td>
							<td>$<?php echo esc_html( number_format( $order_summary['package']['total'], 2 ) ); ?></td>
						</tr>
					<?php foreach ( $order_summary['addOns'] as $addon ) : ?>
							<tr>
								<td><?php echo esc_html( $addon['name'] ); ?></td>
								<td><?php echo esc_html( $addon['quantity'] ); ?></td>
								<td>$<?php echo esc_html( number_format( $addon['price'], 2 ) ); ?></td>
								<td>$<?php echo esc_html( number_format( $addon['price'] * $addon['quantity'], 2 ) ); ?></td>
							</tr>
						<?php endforeach; ?>
					<?php
					// Update the tax row in the email template
					if ( isset( $order_summary['tax'] ) && $order_summary['tax'] > 0 ) {
						?>
							<tr>
								<td colspan="3" class="pt-4 font-semibold text-right">
							<?php
							esc_html_e( 'Tax (', 'serviceez' );
							echo esc_html( $order_summary['taxRate'] );
							esc_html_e( '%)', 'serviceez' );
							?>
								</td>
								<td class="pt-4 text-right font-semibold">$<?php echo esc_html( number_format( $order_summary['tax'], 2 ) ); ?></td>
							</tr>
							<?php
					}
					?>
					</tbody>
					<tfoot>
						<tr class="total">
							<td colspan="3"><?php esc_html_e( 'Total', 'serviceez' ); ?></td>
							<td>$<?php echo esc_html( number_format( $order_summary['total'], 2 ) ); ?></td>
						</tr>
					</tfoot>
				</table>
				<p>
					<a href="<?php echo esc_url( $invoice_link ); ?>" class="button">View Invoice</a>
				</p>
				<p><?php esc_html_e( 'Thank you for your business!', 'serviceez' ); ?></p>
			</div>
		</body>

		</html>
			<?php
			return ob_get_clean();
	}

	/**
	 * Get all orders.
	 *
	 * @return WP_REST_Response The response containing all orders.
	 */
	public function get_all( WP_REST_Request $request ) {
		$page     = $request->get_param( 'page' ) ?? 1;
		$per_page = $request->get_param( 'per_page' ) ?? 20;

		// If user is not admin, return only orders for the current user.
		if ( ! current_user_can( 'administrator' ) ) {
			$user_id = $request->get_header( 'X-User-Id' );

			$orders = Order::where( 'client_id', $user_id )->get();

		} else {
			$orders = Order::with( array( 'client', 'service' ) )
			->skip( ( $page - 1 ) * $per_page )
				->take( $per_page )
				->get();
		}
		return new WP_REST_Response( $orders );
	}

	/**
	 * Get a single order by ID.
	 *
	 * @param WP_REST_Request $request The request object containing the order ID.
	 * @return WP_REST_Response The response containing the order or an error message.
	 */
	public function get_single( $request ) {
		$id = $request->get_param( 'id' );

		if ( ! is_numeric( $id ) ) {
			return new WP_REST_Response( array( 'message' => 'Invalid order ID' ), 400 );
		}

		$order = Order::with( array( 'client', 'service' ) )->find( (int) $id );
		if ( ! $order ) {
			return new WP_REST_Response( array( 'message' => 'Order not found' ), 404 );
		}

		// Decode the JSON data stored in the order.
		$order_data = json_decode( $order->data, true );

		// Ensure $order_data is an array.
		$order_data = is_array( $order_data ) ? $order_data : array();

		// Transform the order data to match the frontend expectations.
		$transformed_order = array(
			'id'               => $order->id,
			'client'           => $order->client,
			'service'          => array(
				'id'          => $order->service->id,
				'name'        => $order->service->name,
				'description' => $order->service->description,
				'price'       => isset( $order_data['package_price'] ) ? $order_data['package_price'] : $order->service->price,
			),
			'status'           => $order->order_status,
			'payment_status'   => $order->payment_status,
			'payment_method'   => $order->payment_method,
			'total_amount'     => $order->total_amount,
			'created_at'       => $order->created_at,
			'billing_info'     => array(
				'firstName'     => $order->client->firstName ?? $order_data['firstName'] ?? '',
				'lastName'      => $order->client->lastName ?? $order_data['lastName'] ?? '',
				'email'         => $order->client->email ?? $order_data['email'] ?? '',
				'phone'         => $order->client->phone ?? $order_data['phone'] ?? '',
				'address_line1' => $order->client->address_line1 ?? $order_data['address']['line1'] ?? '',
				'address_line2' => $order->client->address_line2 ?? $order_data['address']['line2'] ?? '',
				'city'          => $order->client->city ?? $order_data['address']['city'] ?? '',
				'state'         => $order->client->state ?? $order_data['address']['state'] ?? '',
				'postal_code'   => $order->client->postal_code ?? $order_data['address']['postalCode'] ?? '',
				'country'       => $order->client->country ?? $order_data['address']['country'] ?? '',
			),
			'package_quantity' => isset( $order_data['package_quantity'] ) ? $order_data['package_quantity'] : 1,
			'add_ons'          => array(),
			'order_data'       => $order_data,
		);

		// Add add-ons information.
		if ( isset( $order_data['add_ons'] ) && is_array( $order_data['add_ons'] ) ) {
			foreach ( $order_data['add_ons'] as $addon ) {
				if ( is_array( $addon ) ) {
					$transformed_order['add_ons'][] = array(
						'id'          => isset( $addon['id'] ) ? $addon['id'] : '',
						'name'        => isset( $addon['name'] ) ? $addon['name'] : 'Unknown Add-on',
						'description' => isset( $addon['description'] ) ? $addon['description'] : '',
						'price'       => isset( $addon['price'] ) ? $addon['price'] : 0,
						'quantity'    => isset( $addon['quantity'] ) ? $addon['quantity'] : 1,
					);
				}
			}
		}

		// Add Stripe-related information if available.
		if ( isset( $order_data['stripe_payment_intent_id'] ) ) {
			$transformed_order['stripe_payment_intent_id'] = $order_data['stripe_payment_intent_id'];
		}
		if ( isset( $order_data['stripe_invoice_id'] ) ) {
			$transformed_order['stripe_invoice_id'] = $order_data['stripe_invoice_id'];
		}

		if ( isset( $order_data['add_ons'] ) ) {
			$transformed_order['add_ons'] = $order_data['add_ons'];
		}

		return new WP_REST_Response( $transformed_order );
	}

	/**
	 * Update an existing order.
	 *
	 * @param WP_REST_Request $request The incoming request object.
	 * @return WP_REST_Response The response containing the updated order or an error message.
	 */
	public function update( $request ) {
		$id = $request->get_param( 'id' );

		if ( ! is_numeric( $id ) ) {
			return new WP_REST_Response( array( 'message' => 'Invalid order ID' ), 400 );
		}

		$order = Order::find( (int) $id );
		if ( ! $order ) {
			return new WP_REST_Response( array( 'message' => 'Order not found' ), 404 );
		}

		$params = $request->get_params();

		// Check if service_id is being updated and if it exists.
		if ( isset( $params['service_id'] ) ) {
			$service = Service::find( $params['service_id'] );
			if ( ! $service ) {
				return new WP_REST_Response( array( 'message' => 'Invalid service ID' ), 400 );
			}
		}

		$order->update( $params );
		$order->load( array( 'client', 'service' ) );
		return new WP_REST_Response( $order );
	}

	/**
	 * Delete an order.
	 *
	 * @param WP_REST_Request $request The request object containing the order ID.
	 * @return WP_REST_Response The response containing a success message or an error message.
	 */
	public function delete( $request ) {
		$id = $request->get_param( 'id' );

		if ( ! is_numeric( $id ) ) {
			return new WP_REST_Response( array( 'message' => 'Invalid order ID' ), 400 );
		}

		$order = Order::find( (int) $id );
		if ( ! $order ) {
			return new WP_REST_Response( array( 'message' => 'Order not found' ), 404 );
		}
		$order->delete();
		return new WP_REST_Response( array( 'message' => 'Order deleted successfully' ) );
	}

	/**
	 * Update the status of an order.
	 *
	 * @param WP_REST_Request $request The request object containing the order ID and status.
	 * @return WP_REST_Response The response containing a success message or an error message.
	 */
	public function update_order_status( $request ) {
		$id                  = $request->get_param( 'id' );
		$status              = $request->get_param( 'status' );
		$order               = Order::find( (int) $id );
		$order->order_status = $status;
		$order->save();

		// Create notification.
		$this->create_notification(
			// translators: %1$d: Order ID, %2$s: Order status.
			sprintf( __( 'Order #%1$d status updated to %2$s', 'serviceez' ), $order->id, $status ),
			'order_status_updated',
			"/orders/{$order->id}"
		);

		return new WP_REST_Response( array( 'message' => 'Order status updated successfully' ) );
	}

	/**
	 * Process payment for an order.
	 *
	 * @param WP_REST_Request $request The request object containing the order ID and payment intent ID.
	 * @return WP_REST_Response The response containing a success message or an error message.
	 * @throws \Exception When payment processing fails.
	 */
	public function process_payment( WP_REST_Request $request ) {
		$order_id          = $request->get_param( 'id' );
		$payment_intent_id = $request->get_param( 'payment_intent_id' );

		$order = Order::find( (int) $order_id );
		if ( ! $order ) {
			return new WP_REST_Response( array( 'message' => 'Order not found' ), 404 );
		}

		\Stripe\Stripe::setApiKey( $this->get_stripe_secret_key() );

		try {
			$payment_intent = \Stripe\PaymentIntent::retrieve( $payment_intent_id );

			if ( 'succeeded' === $payment_intent->status ) {
				$this->update_order_after_payment( $order, $payment_intent );
				return new WP_REST_Response(
					array(
						'message' => 'Payment processed successfully',
						'order'   => $order,
					)
				);
			} else {
				$order->payment_status = 'failed';
				$order->save();
				throw new \Exception( 'Payment failed' );
			}
		} catch ( \Exception $e ) {
			return new WP_REST_Response( array( 'message' => $e->getMessage() ), 400 );
		}
	}

	/**
	 * Update order after successful payment.
	 *
	 * @param Order                 $order The order to update.
	 * @param \Stripe\PaymentIntent $intent The Stripe PaymentIntent object.
	 */
	private function update_order_after_payment( $order, $intent ) {
		$order->order_status   = 'completed';
		$order->payment_status = 'paid';
		$order->payment_method = 'stripe';
		$order->save();
	}

	/**
	 * Get the Stripe secret key.
	 *
	 * @return string The Stripe secret key.
	 */
	private function get_stripe_secret_key() {
		$settings = get_option( 'serviceez_settings', array() );
		return $settings['payment']['stripeSecretKey'] ?? '';
	}

	/**
	 * Confirm a payment that required additional action.
	 *
	 * @param WP_REST_Request $request The request object containing the order ID and payment intent ID.
	 * @return WP_REST_Response The response containing a success message or an error message.
	 * @throws \Exception When payment confirmation fails.
	 */
	public function confirm_payment( $request ) {
		$order_id          = $request->get_param( 'id' );
		$payment_intent_id = $request->get_param( 'payment_intent_id' );

		$order = Order::find( (int) $order_id );
		if ( ! $order ) {
			return new WP_REST_Response( array( 'message' => 'Order not found' ), 404 );
		}

		Stripe::setApiKey( $this->get_stripe_secret_key() );

		try {
			$intent = PaymentIntent::retrieve( $payment_intent_id );

			// Confirm the PaymentIntent without updating the payment method.
			$intent->confirm();

			if ( 'succeeded' === $intent->status ) {
				$this->update_order_after_payment( $order, $intent );
				return new WP_REST_Response(
					array(
						'message' => 'Payment confirmed successfully',
						'order'   => $order,
					)
				);
			} elseif ( 'requires_action' === $intent->status ) {
				$order->payment_status = 'pending';
				$order->save();
				return new WP_REST_Response(
					array(
						'requires_action' => true,
						'client_secret'   => $intent->client_secret,
					),
					200
				);
			} else {
				$order->payment_status = 'failed';
				$order->save();
				throw new \Exception( 'Payment confirmation failed' );
			}
		} catch ( \Exception $e ) {
			return new WP_REST_Response( array( 'message' => $e->getMessage() ), 400 );
		}
	}

	/**
	 * Create a notification.
	 *
	 * @param string   $message The notification message.
	 * @param string   $type The notification type.
	 * @param string   $link The notification link.
	 * @param int|null $user_id The user ID.
	 */
	private function create_notification( $message, $type, $link, $user_id = null ) {
		Notification::create(
			array(
				'user_id' => $user_id ?? get_current_user_id(),
				'message' => $message,
				'type'    => $type,
				'link'    => $link,
			)
		);
	}

	/**
	 * Send an update email to the customer.
	 *
	 * @param WP_REST_Request $request The request object containing the email details.
	 * @return WP_REST_Response The response containing success or error message.
	 */
	public function send_update_email( WP_REST_Request $request ) {
		$order_id = $request->get_param( 'id' );
		$subject  = sanitize_text_field( $request->get_param( 'subject' ) );
		$message  = wp_kses_post( $request->get_param( 'message' ) );

		$order = Order::with( 'client' )->find( (int) $order_id );
		if ( ! $order ) {
			return new WP_REST_Response( array( 'message' => 'Order not found' ), 404 );
		}

		try {
			// Send email using the EmailHelper.
			$sent = EmailHelper::send_email(
				$order->client->email,
				$subject,
				$message
			);

			if ( $sent ) {
				// Create notification for the update.
				$this->create_notification(
					// translators: %d: Order ID.
					sprintf( __( 'Update email sent for Order #%d', 'serviceez' ), $order->id ),
					'order_update_sent',
					"/orders/{$order->id}"
				);

				return new WP_REST_Response(
					array(
						'message' => 'Update email sent successfully',
						'order'   => $order,
					)
				);
			}

			throw new \Exception( 'Failed to send email' );
		} catch ( \Exception $e ) {
			return new WP_REST_Response(
				array( 'message' => 'Failed to send update email: ' . $e->getMessage() ),
				500
			);
		}
	}
}

