<?php
/**
 * Service Actions Controller
 *
 * Handles creating, editing, and deleting services.
 *
 * @package ServiceEZ\Controllers\Services
 * @since 1.0.0
 */

namespace ServiceEZ\Controllers\Services;

use ServiceEZ\Models\Service;
use WP_REST_Request;
use WP_REST_Response;
use WP_Error;

/**
 * Actions class for handling service-related operations.
 */
class Actions {

	/**
	 * Create a new service.
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response|WP_Error The response object or WP_Error on failure.
	 */
	public function create( WP_REST_Request $request ) {
		$params = $request->get_params();

		// Validate required fields.
		$required_fields = array( 'service_name', 'data' );
		foreach ( $required_fields as $field ) {
			if ( empty( $params[ $field ] ) ) {
				return new WP_Error( 'missing_field', "Missing required field: {$field}", array( 'status' => 400 ) );
			}
		}

		// Validate that data contains serviceTypes.
		$data = json_decode( $params['data'], true );
		if ( ! isset( $data['serviceTypes'] ) || empty( $data['serviceTypes'] ) ) {
			return new WP_Error( 'invalid_data', "The 'data' field must contain 'serviceTypes'", array( 'status' => 400 ) );
		}

		$service_data = json_decode( $params['data'], true );

		$params['data'] = wp_json_encode( Format::service_data( $service_data ) );

		// Create new service.
		try {
			$service = Service::create( $params );
			return new WP_REST_Response( $service, 201 );
		} catch ( \Exception $e ) {
			return new WP_Error( 'create_failed', $e->getMessage(), array( 'status' => 500 ) );
		}
	}

	/**
	 * Edit an existing service.
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response|WP_Error The response object or WP_Error on failure.
	 */
	public function edit( WP_REST_Request $request ) {
		$service_id = $request->get_param( 'id' );
		$params     = $request->get_params();

		// Check if service exists.
		$service = Service::where( 'id', $service_id )->first();
		if ( ! $service ) {
			return new WP_Error( 'not_found', 'Service not found', array( 'status' => 404 ) );
		}

		// Update service.
		try {
			$service->update( $params );
			return new WP_REST_Response( $service, 200 );
		} catch ( \Exception $e ) {
			return new WP_Error( 'update_failed', $e->getMessage(), array( 'status' => 500 ) );
		}
	}

	/**
	 * Delete a service.
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response|WP_Error The response object or WP_Error on failure.
	 */
	public function delete( WP_REST_Request $request ) {
		$service_id = $request->get_param( 'id' );

		// Check if service exists.
		$service = Service::where( 'id', $service_id )->first();
		if ( ! $service ) {
			return new WP_Error( 'not_found', 'Service not found', array( 'status' => 404 ) );
		}

		// Delete service.
		try {
			$service->delete();
			return new WP_REST_Response( null, 204 );
		} catch ( \Exception $e ) {
			return new WP_Error( 'delete_failed', $e->getMessage(), array( 'status' => 500 ) );
		}
	}

	/**
	 * Get all services.
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response|WP_Error The response object or WP_Error on failure.
	 */
	public function get_all( WP_REST_Request $request ) {
		try {
			$services           = Service::all();
			$formatted_services = $services->map(
				function ( $service ) {
					$data = json_decode( $service['data'], true );
					return array(
						'id'            => $service['id'],
						'service_name'  => $service['service_name'],
						'service_type'  => $service['service_type'],
						'service_types' => isset( $data['serviceTypes'] ) && is_array( $data['serviceTypes'] )
						? array_map(
							function ( $service_type ) {
								return $service_type['name'];
							},
							$data['serviceTypes']
						)
						: array(),
						'addons'        => isset( $data['addOns'] ) && is_array( $data['addOns'] )
						? array_reduce(
							$data['addOns'],
							function ( $carry, $addons ) {
								return array_merge(
									$carry,
									array_map(
										function ( $addon ) {
											return $addon['name'];
										},
										$addons
									)
								);
							},
							array()
						)
						: array(),
						'data'          => $data,
						'created_at'    => $service['created_at'],
					);
				}
			)->toArray();

			return new WP_REST_Response( $formatted_services, 200 );
		} catch ( \Exception $e ) {
			return new WP_Error( 'fetch_failed', $e->getMessage(), array( 'status' => 500 ) );
		}
	}

	/**
	 * Get a single service.
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response|WP_Error The response object or WP_Error on failure.
	 */
	public function get_single( WP_REST_Request $request ) {
		$service_id = $request->get_param( 'id' );

		try {
			$service = Service::where( 'id', $service_id )->first();
			if ( ! $service ) {
				return new WP_Error( 'not_found', 'Service not found', array( 'status' => 404 ) );
			}
			return new WP_REST_Response( $service, 200 );
		} catch ( \Exception $e ) {
			return new WP_Error( 'fetch_failed', $e->getMessage(), array( 'status' => 500 ) );
		}
	}

	/**
	 * Get share service data.
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response|WP_Error The response object or WP_Error on failure.
	 */
	public function get_share_service( WP_REST_Request $request ) {
		$service_id = $request->get_param( 'id' );

		try {
			$service = Service::where( 'id', $service_id )->first();
			if ( ! $service ) {

				return new WP_Error( 'not_found', 'Service not found', array( 'status' => 404 ) );
			}

			// Format the service data for the frontend.
			$data              = json_decode( $service['data'], true );
			$formatted_service = array(
				'id'           => $service['id'],
				'service_name' => $service['service_name'],
				'config'       => $data,
			);

			return new WP_REST_Response( $formatted_service, 200 );
		} catch ( \Exception $e ) {

			return new WP_Error( 'fetch_failed', $e->getMessage(), array( 'status' => 500 ) );
		}
	}
}
