<?php
/**
 * Settings Actions
 *
 * @package ServiceEZ\Controllers\Settings
 */

namespace ServiceEZ\Controllers\Settings;

use ServiceEZ\Models\Client;
use WP_REST_Request;
use WP_REST_Response;

// Add WordPress function imports.
use function get_option;
use function update_option;
use function get_user_by;
use function wp_update_user;
use function absint;


/**
 * Settings Actions
 *
 * @package ServiceEZ\Controllers\Settings
 */
class Actions {
	/**
	 * Get settings
	 *
	 * @return WP_REST_Response
	 */
	public function get_settings( WP_REST_Request $request ) {
		$default_settings = array(
			'general' => array(
				'companyName' => '',
				'language'    => '',
				'timezone'    => '',
				'dateFormat'  => '',
				'currency'    => '',
			),
			'user'    => array(
				'theme'              => '',
				'emailNotifications' => false,
				'pushNotifications'  => false,
				'twoFactorAuth'      => false,
			),
			'api'     => array(
				'apiKey'             => '',
				'webhookUrl'         => '',
				'rateLimitPerMinute' => 60,
			),
			'payment' => array(
				'stripePublishableKey' => '',
				'stripeSecretKey'      => '',
			),
		);

		$saved_settings = get_option( 'serviceez_settings', array() );
		$settings       = array_replace_recursive( $default_settings, $saved_settings );
		if ( ! $request->get_header( 'X-Settings-Page' ) ) {
			unset( $settings['payment']['stripeSecretKey'] );
		}
		return new WP_REST_Response( $settings );
	}

	/**
	 * Update settings
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response
	 */
	public function update_settings( WP_REST_Request $request ) {
		$nonce    = $request->get_header( 'X-WP-Nonce' );
		$settings = $request->get_json_params();

		if ( ! wp_verify_nonce( $nonce, 'wp_rest' ) ) {
			return new WP_REST_Response( array( 'message' => 'Invalid nonce' ), 403 );
		}

		if ( $request->get_param( 'is_onboarding' ) ) {
			update_option( 'serviceez_onboarding', true );
			// Sanitize settings recursively.
			$sanitized_settings = $this->sanitize_settings_recursive( $settings );
			// Update the settings.
			update_option( 'serviceez_settings', $sanitized_settings );
			return new WP_REST_Response( array( 'message' => 'Settings updated successfully' ), 200 );
		}

		try {

			// Validate required fields.
			if ( empty( $settings ) ) {
				return new WP_REST_Response(
					array( 'message' => 'No settings provided' ),
					400
				);
			}

			// Sanitize settings recursively.
			$sanitized_settings = $this->sanitize_settings_recursive( $settings );

			// Update the settings.
			$updated = update_option( 'serviceez_settings', $sanitized_settings );
			if ( ! $updated ) {
				return new WP_REST_Response(
					array( 'message' => 'Failed to update settings' ),
					500
				);
			}

			return new WP_REST_Response(
				array(
					'message'  => 'Settings updated successfully',
					'settings' => $sanitized_settings,
				)
			);

		} catch ( Exception $e ) {
			return new WP_REST_Response(
				array( 'message' => $e->getMessage() ),
				500
			);
		}
	}

	private function sanitize_settings_recursive( $settings ) {
		$sanitized = array();

		foreach ( $settings as $key => $value ) {
			if ( is_array( $value ) ) {
				$sanitized[ $key ] = $this->sanitize_settings_recursive( $value );
			} else {
				$sanitized[ $key ] = sanitize_text_field( $value );
			}
		}

		return $sanitized;
	}

	/**
	 * Update profile
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response
	 */
	public function update_profile( WP_REST_Request $request ) {
		$user_id = absint( $request->get_header( 'X-User-ID' ) );
		$client  = Client::where( 'user_id', $user_id )->first();

		if ( $client ) {
			$profile_data = array_map( 'sanitize_text_field', $request->get_json_params()['profile'] );
			$client->update( $profile_data );

			// Check if WordPress user exists.
			$user = get_user_by( 'ID', $user_id );
			if ( $user ) {
				// Update WordPress user password if password is not empty.
				if ( ! empty( $request->get_json_params()['profile']['password'] ) ) {
					wp_update_user(
						array(
							'ID'        => $user_id,
							'user_pass' => $request->get_json_params()['profile']['password'],
						)
					);
				}
			}

			return new WP_REST_Response( array( 'message' => 'Profile updated successfully' ) );
		}

		return new WP_REST_Response( array( 'message' => 'Client not found' ), 404 );
	}
}
