<?php
/**
 * Stripe Actions Controller
 *
 * This file contains the Actions class for handling Stripe-related operations.
 *
 * @package ServiceEZ
 * @subpackage Controllers\Stripe
 */

namespace ServiceEZ\Controllers\Stripe;

use ServiceEZ\Models\Order;
use WP_REST_Request;
use WP_REST_Response;
use Stripe\Stripe;

/**
 * Class Actions
 *
 * Handles Stripe payment operations.
 */
class Actions {
	/**
	 * Get the Stripe secret key.
	 *
	 * @return string The Stripe secret key.
	 */
	private function get_stripe_secret_key() {
		$settings = get_option( 'serviceez_settings', array() );
		return $settings['payment']['stripeSecretKey'] ?? '';
	}

	/**
	 * Create a PaymentIntent for Stripe.
	 *
	 * @param WP_REST_Request $request The incoming request object.
	 * @return WP_REST_Response The response containing the PaymentIntent client secret.
	 */
	public function create_payment_intent( WP_REST_Request $request ) {
		$params = $request->get_params();

		try {
			$secret_key = $this->get_stripe_secret_key();
			if ( empty( $secret_key ) ) {
				throw new \Exception( 'Stripe secret key is not set' );
			}

			\Stripe\Stripe::setApiKey( $secret_key );
		

			if ( isset( $params['type'] ) && 'invoice' === $params['type'] ) {
				$params = $params['order_data']['order_data'];
			}

			// Create or get customer.
			$customer = $this->get_or_create_customer( $params['customerEmail'] );

			// Handle recurring subscription.
			if ( $this->is_recurring_payment( $params ) ) {

				$subscription         = new Subscription( $customer );
				$created_subscription = $subscription->create( $params );

				if ( $created_subscription ) {
					return new WP_REST_Response(
						array(
							'clientSecret'   => $created_subscription->latest_invoice->payment_intent->client_secret,
							'subscriptionId' => $created_subscription->id,
						)
					);
				}
			} else {
				// Handle one-time payment.

				$payment_intent         = new Invoice( $customer );
				$created_payment_intent = $payment_intent->create( $params );
			
				return new WP_REST_Response(
					array(
						'clientSecret'    => $created_payment_intent->client_secret,
						'paymentIntentId' => $created_payment_intent->id,
					)
				);

			}
		} catch ( \Exception $e ) {

			return new WP_REST_Response( array( 'error' => $e->getMessage() ), 400 );
		}
	}

	/**
	 * Get or create a Stripe customer.
	 *
	 * @param string $email The customer's email address.
	 * @return \Stripe\Customer The Stripe customer object.
	 */
	private function get_or_create_customer( $email ) {
		try {
			$existing_customers = \Stripe\Customer::search(
				array(
					'query' => "email:'$email'",
					'limit' => 1,
				)
			);

			if ( ! empty( $existing_customers->data ) ) {
				return $existing_customers->data[0];
			}

			$customer = \Stripe\Customer::create(
				array(
					'email' => $email,
				)
			);
			return $customer;
		} catch ( \Exception $e ) {

			throw $e;
		}
	}

	/**
	 * Get an invoice by ID.
	 *
	 * @param WP_REST_Request $request The incoming request object.
	 * @return WP_REST_Response The response containing the invoice.
	 */
	public function get_invoice( WP_REST_Request $request ) {
		try {
			\Stripe\Stripe::setApiKey( $this->get_stripe_secret_key() );
			$invoice_id = $request->get_param( 'invoice_id' );
			$invoice    = \Stripe\Invoice::retrieve( $invoice_id );

			return new WP_REST_Response(
				array(
					'invoice' => $invoice,
				)
			);
		} catch ( \Exception $e ) {
			return new WP_REST_Response( array( 'error' => $e->getMessage() ), 400 );
		}
	}

	public function is_recurring_payment( $params ) {

		$selected_packages = $params['selectedPackage'];
		$selected_addons   = $params['selectedAddOns'];

		if ( ! empty( $selected_packages ) && 'recurring' === $selected_packages['pricingType'] ) {
			return true;
		}
		if ( empty( $selected_addons ) ) {
			return false;
		}
		foreach ( $selected_addons as $addon ) {

			if ( ! empty( $addon['pricingType'] ) && 'recurring' === $addon['pricingType'] ) {
				return true;
			}
		}
		return false;
	}
}
