<?php
/**
 * Handles Stripe invoice operations
 *
 * @package ServiceEZ
 * @subpackage Controllers\Stripe
 */

namespace ServiceEZ\Controllers\Stripe;

/**
 * Invoice class for managing Stripe invoice operations
 */
class Invoice {

	/**
	 * Stripe Customer instance
	 *
	 * @var \Stripe\Customer
	 */
	private $customer;

	/**
	 * Constructor
	 *
	 * @param \Stripe\Customer $customer Stripe customer instance.
	 */
	public function __construct( $customer ) {
		$this->customer = $customer;
	}

	/**
	 * Creates a payment intent for the invoice
	 *
	 * @param array $params Parameters for creating the payment intent.
	 * @return \Stripe\PaymentIntent|false Payment intent object or false on failure
	 */
	public function create( $params ) {
		try {
			$args =
				array(

					'customer' => $this->customer->id,

					'metadata' => array(
						'package_id'   => $params['selectedPackage']['package_id'],
						'service_name' => $params['selectedPackage']['service_name'],
					),

				);

			$invoice = \Stripe\Invoice::create( $args );

			// Add invoice item.
			$selected_package = $params['selectedPackage'];

			$selected_package_price_data = array(
				'customer'    => $this->customer->id,
				'description' => $selected_package['description'],
				'currency'    => 'usd',
				'invoice'     => $invoice->id,
				'quantity'    => isset( $selected_package['quantity'] ) ? $selected_package['quantity'] : 1,
				'unit_amount' => $selected_package['price'] * 100,
			);

			if ( ! empty( $selected_package['price_id'] ) ) {
				$selected_package_price_data['price'] = $selected_package['price_id'];
				unset( $selected_package_price_data['unit_amount'] );
			}

			\Stripe\InvoiceItem::create( $selected_package_price_data );

			foreach ( $params['selectedAddOns'] as $addon ) {
				$addon_price_data = array(
					'customer'    => $this->customer->id,
					'quantity'    => isset( $addon['quantity'] ) ? $addon['quantity'] : 1,
					'description' => $addon['description'],
					'currency'    => 'usd',
					'invoice'     => $invoice->id,
					'unit_amount' => $addon['price'] * 100,
				);

				if ( ! empty( $addon['price_id'] ) ) {
					$addon_price_data['price'] = $addon['price_id'];
					unset( $addon_price_data['unit_amount'] );
				}

				\Stripe\InvoiceItem::create( $addon_price_data );

			}

			$invoice = $invoice->finalizeInvoice();

			// Retrieve the payment intent.
			$payment_intent = \Stripe\PaymentIntent::retrieve( $invoice->payment_intent );

			return $payment_intent;
		} catch ( \Exception $e ) {
			
			return false;
		}
	}
}
